PROGRAM Generate_Pkz
  USE cosmo
  USE growth
  USE linearpk
  USE sigma
  USE sigmagaus
  ! A program for generating 2-dimensional arrays of non-linear 
  ! power spectra, P(k,z), as a function of wavenumbers k and redshifts z. 
  ! It uses 3 different methods for computing P(k,z):
  ! 1) Peacock and Dodds [Peacock and Dodds, MNRAS, 280, L19 (1996)]
  ! 2) Ma et al. [Ma et al., ApJL, 521, L1 (1999)]
  ! 3) Halofit [Smith et al., MNRAS, 341, 1311 (2003)]
  ! - k is in units of h Mpc^-1
  ! - P(k) is in units of h^-3 Mpc^3
  ! This code will generate large files containing P(k,z)  as a 
  ! function of both k AND z.
  ! January 1, 2012: E.Komatsu
  IMPLICIT none
  character(len=128) :: filename
  integer :: n
  real :: chebev,lnsigma2
  double precision :: g,z,dz,D,sig8,pi=3.14159d0
  double precision :: linear_pk,k_ov_h,pk,dlnk,lnk,zin
  external g,linear_pk
  double precision :: D2nl,kNL
  ! For Peacock&Dodds(1996)
  double precision :: dlnpkdlnk
  double precision :: y,A,B,alpha,beta,V 
  external dlnpkdlnk
  ! For Ma et al. (1999)
  double precision :: g0,c1,c2 
  ! For Halofit; Smith et al. (2003)
  real :: dlnsigma2dlnRh,lnRh
  double precision :: rneff,rncur,rknl
  integer :: i,iter,imax=50 
  double precision, allocatable :: Rtest(:)
  double precision :: zbrent,func
  double precision :: ltest,m1,m2,ml,mu
  external zbrent,func
  ! For interpolating PD96 and Ma99 at the wavenumbers of Halofit
  integer :: nk=691 ! # of data points in wavenumber
  integer :: nz=901 ! # of data points in redshift
  integer :: ik,iz
  double precision :: lnPk
  double precision, allocatable, dimension(:) :: x1,x2,x3,y1,y2,z1,z2
! Specify three cosmological parameters
! The data type has been defined in MODULE cosmo.
  ode0=0.723d0
  om0=0.277d0
  w=-1d0
! read in and tabulate P(k)
  filename='wavenumber_pknowiggle_at_z=30.txt'
  !!$filename='wmap5baosn_max_likelihood_matterpower_at_z=30.dat'
  n=896 ! # of lines in the file
  zin=30d0
  CALL open_linearpk(filename,n)
! compute and fit sigma^2(R) to Chebyshev polynomials
  CALL compute_sigma2
  lnsigma2 = CHEBEV(lnR1,lnR2,c,ndim,alog(8.)) ! with the tophat window
! For Halofit, compute and fit sigma^2(R) WITH A GAUSSIAN WINDOW to 
! Chebyshev polynomials
  CALL compute_sigma2gaus
! tabulate g(z) by calling "setup_growth"
  call setup_growth
! output P(k,z)
  open(1,file='redshift_wavenumber_pkpd96.txt',status='unknown')
  open(2,file='redshift_wavenumber_pkma99.txt',status='unknown')
  open(3,file='redshift_wavenumber_pkhalofit.txt',status='unknown')
! Loop over redshifts
  z=0d0
  dz=1d-2
  do iz=1,nz
     D=g(z)/g(zin)*(1d0+zin)/(1d0+z) ! linear growth relative to zin
     ! compute sigma8 at a given redshift
     sig8=dble(exp(0.5*lnsigma2))*D
     !!$print*,'sigma8 at z=',z,'is',sig8
     ! For Halofit, find initial radius range for zbrent 
     ! (non-lin. equation solver), within which Rsigma will be found.
     allocate(Rtest(0:imax))
     ! initial guess for Rsigma [h^-1 Mpc]
     Rtest(0)=3d0 
     ltest=func(dlog(Rtest(0)),D)
     if(ltest.le.0d0)then
        do i=1,imax
           Rtest(i)=Rtest(i-1)/2.d0
           ltest=func(dlog(Rtest(i)),D)
           if (ltest.gt.0d0)then
              m1=dlog(Rtest(i))
              m2=dlog(Rtest(i-1))
              exit
           endif
        enddo
     else
        do i=1,imax
           Rtest(i)=Rtest(i-1)*2.d0
           ltest=func(dlog(Rtest(i)),D)
           if(ltest.lt.0d0)then
              m1=dlog(Rtest(i))
              m2=dlog(Rtest(i-1))
              exit
           endif
        enddo
     endif
     deallocate(Rtest)
     ! Lower & upper limits for R, between which Rsigma will be found.
     ml=min(m1,m2) 
     mu=max(m1,m2)
     ! Find Rsigma [h^-1 Mpc] satisfying Sigma(Rsigma)=1 using zbrent.
     lnRh=real(zbrent(func,ml,mu,1d-4,iter,D))
     dlnsigma2dlnRh=CHEBEV(lnR1g,lnR2g,cgder,ndimg,lnRh) ! dlnSigma^2/dlnR
     rneff=-3d0-dble(dlnsigma2dlnRh)
     rncur=-dble(CHEBEV(lnR1g,lnR2g,cgderder,ndimg,lnRh)) ! -d^2lnSigma^2/dlnR^2
     rknl=dble(exp(-lnRh))
     !!$print*,'Halofit: effective spectral quantities:'
     !!$write(*,20) 'rknl [h/Mpc] =',rknl,'rneff=',rneff, 'rncur=',rncur
20   format(a14,f12.6,2x,a6,f12.6,2x,a6,f12.6)
     ! Loop over wavenumbers 
     k_ov_h=1d-4 ! h/Mpc
     dlnk=2d-2
     ALLOCATE(x1(nk),x2(nk),x3(nk),y1(nk),y2(nk),z1(nk),z2(nk))
     do ik=1,nk
        !------------------- Peacock & Dodds (1996) -------------------
        y=1d0+dlnpkdlnk(0.5d0*k_ov_h)/3d0
        A=0.482d0*y**(-0.947d0)
        B=0.226d0*y**(-1.778d0)
        alpha=3.310d0*y**(-0.244d0)
        beta=0.862d0*y**(-0.287d0)
        V=11.55d0*y**(-0.423d0)
        y=k_ov_h**3d0*D**2d0*linear_pk(k_ov_h)/(2d0*pi*pi) ! L k^3*P(k)/(2*pi^2)
        D2nl=y*((1d0+B*beta*y+(A*y)**(alpha*beta)) & ! NL  k^3*P(k)/(2*pi^2)
             /(1d0+((A*y)**alpha*g(z)**3d0/(V*y**0.5))**beta))**(1d0/beta)
        kNL=k_ov_h*(1d0+D2nl)**(1d0/3d0) ! non-linear wavenumber
        pk=D2nl*(2d0*pi*pi)/kNL**3d0 ! non-linear P(k)
        x1(ik)=dlog(kNL)
        y1(ik)=dlog(pk)
        !------------------- Ma et al. (1999) -------------------
        beta=0.83d0
        c1=1.08d-4
        c2=2.10d-5
        g0=g(0d0)*dabs(w)**(1.3d0*dabs(w)-0.76d0)
        ! linear k^3*P(k)/(2*pi^2) divided by g0^1.5*sig8(z)^beta
        y=k_ov_h**3d0*D**2d0*linear_pk(k_ov_h)/(2d0*pi*pi)/g0**1.5d0/sig8**beta
        D2nl=k_ov_h**3d0*D**2d0*linear_pk(k_ov_h)/(2d0*pi*pi)&!NL k^3*P(k)/(2*pi^2)
             *(1d0+dlog(1d0+0.5d0*y))*(1d0+0.02d0*y**4d0+c1*y**8d0/g(z)**3d0) &
             /(1d0+c2*y**7.5d0)
        kNL=k_ov_h*(1d0+D2nl)**(1d0/3d0) ! non-linear wavenumber
        pk=D2nl*(2d0*pi*pi)/kNL**3d0 ! non-linear P(k)
        x2(ik)=dlog(kNL)
        y2(ik)=dlog(pk)
        !------------------ Halofit, Smith et al. (2003) ------------------
        y=k_ov_h**3d0*D**2d0*linear_pk(k_ov_h)/(2d0*pi*pi) ! lin. k^3*P(k)/(2*pi^2)
        CALL halofit(z,k_ov_h,rneff,rncur,rknl,y,D2nl)
        pk=D2nl*(2d0*pi*pi)/k_ov_h**3d0 ! non-linear P(k)
        write(3,'(3E18.8)')z,k_ov_h,pk
        x3(ik)=dlog(k_ov_h)
        lnk=dlog(k_ov_h)+dlnk
        k_ov_h=dexp(lnk)
     enddo
     ! Interpolating PD96 and Ma99 at the wavenumbers of Halofit
     CALL spline(x1,y1,nk,1.d30,1.d30,z1)
     CALL spline(x2,y2,nk,1.d30,1.d30,z2)
     do ik=1,nk
        CALL splint(x1,y1,z1,nk,x3(ik),lnPk)
        write(1,'(3E18.8)')z,dexp(x3(ik)),dexp(lnPk)
        CALL splint(x2,y2,z2,nk,x3(ik),lnPk)
        write(2,'(3E18.8)')z,dexp(x3(ik)),dexp(lnPk)
     enddo
     DEALLOCATE(x1,x2,x3,y1,y2,z1,z2)
     z=z+dz
  enddo
  call close_linearpk
  !!$print*,'the number of data points in wavenumber=',nk,'from k=',dexp(x3(1)),'to',dexp(x3(nk)),'h/Mpc'
  !!$print*,'the number of data points in redshift=',nz,'from z=0 to',z-dz
  close(1)
  close(2)
  close(3)
END PROGRAM Generate_Pkz
!-----------------------------------------------------------------------------
double precision function func(lnRh,growth)
  USE sigmagaus
! zbrent will solve the equation, sigma(Rsigma,z)=1, for ln(Rsigma).
  IMPLICIT none
  double precision, intent(IN) :: lnRh,growth ! Rh in units of h^-1 Mpc
  real :: chebev,lnsigma2
  lnsigma2=CHEBEV(lnR1g,lnR2g,cg,ndimg,real(lnRh))!ln(sigma^2) with Gaussian window
  func=0.5d0*dble(lnsigma2)+dlog(growth)-0d0 
  return
END function func
